/**
 * An ADC client emulator.
 */

#include "uhub.h"

#define ADC_CLIENTS_DEFAULT 1
#define ADC_MAX_CLIENTS 25000

#define ADC_BUFSIZE 16384
#define ADC_SIDSIZE 4
#define ADC_CID_SIZE 39

#define BIG_BUFSIZE 131072
#define TIGERSIZE 24

#define ADC_HANDSHAKE "HSUP ADBASE ADTIGR\n"
#define ADCRUSH "adcrush/0.1"

static struct sockaddr_in saddr;

static char* cfg_host = 0;
static int cfg_port = 0;

static int cfg_bugmode  = 0; /* bug mode -- make "honest" mistakes */
static int cfg_debug    = 0; /* debug level */
static int cfg_level    = 1; /* activity level (0..3) */
static int cfg_chat     = 0; /* chat mode, allow sending chat messages */
static int cfg_fuzz     = 0; /* fuzzer mode, construct weird commands */
static int cfg_quiet    = 0; /* quiet mode (no output) */
static int cfg_security = 0; /* perform security policy tests */
static int cfg_clients  = ADC_CLIENTS_DEFAULT; /* number of clients */

enum commandMode
{
	cm_bcast    = 0x01, /* B - broadcast */
	cm_dir      = 0x02, /* D - direct message */
	cm_echo     = 0x04, /* E - echo message */
	cm_fcast    = 0x08, /* F - feature cast message */
	cm_c2h      = 0x10, /* H - client to hub message */
	cm_h2c      = 0x20, /* I - hub to client message */
	cm_c2c      = 0x40, /* C - client to client message */
	cm_udp      = 0x80, /* U - udp message (client to client) */
};

enum commandValidity
{
	cv_protocol = 0x01,
	cv_identify = 0x02,
	cv_verify   = 0x04,
 	cv_normal   = 0x08,
};

enum protocolState
{
	ps_none     = 0x00, /* none or disconnected */
	ps_protocol = 0x01,
	ps_identify = 0x02,
	ps_verify   = 0x04,
	ps_normal   = 0x08,
};

struct commandPattern
{
	unsigned char mode; /* see enum commandMode */
	char cmd[3];
	unsigned char validity; /* see enum commandValidity */
};

const struct commandPattern patterns[] =
{
	{ cm_c2h | cm_c2c | cm_h2c,                     "SUP", cv_protocol | cv_normal }, /* protocol support */
	{ cm_bcast | cm_h2c | cm_c2c,                   "INF", cv_identify | cv_verify | cv_normal }, /* info message */
	{ cm_bcast | cm_h2c | cm_c2c | cm_c2h | cm_udp, "STA", cv_protocol | cv_identify | cv_verify | cv_normal }, /* status message */
	{ cm_bcast | cm_dir | cm_echo | cm_h2c,         "MSG", cv_normal },   /* chat message */
	{ cm_bcast | cm_dir | cm_echo | cm_fcast,       "SCH", cv_normal },   /* search */
	{ cm_dir | cm_udp,                              "RES", cv_normal },   /* search result */
	{ cm_dir | cm_echo,                             "CTM", cv_normal },   /* connect to me */
	{ cm_dir | cm_echo,                             "RCM", cv_normal },   /* reversed, connect to me */
	{ cm_h2c,                                       "QUI", cv_normal },   /* quit message */
	{ cm_h2c,                                       "GPA", cv_identify }, /* password request */
	{ cm_c2h,                                       "PAS", cv_verify }    /* password response */
};

struct ADC_client
{
	int  sd;
	int  num;
	sid_t sid;
	enum protocolState state;
	char info[ADC_BUFSIZE];
	char recvbuf[BIG_BUFSIZE];
	char sendbuf[BIG_BUFSIZE];
	size_t s_offset;
	size_t r_offset;
	struct event ev_read;
	struct event ev_write;
	struct event ev_timer;
	size_t timeout;
};

static void ADC_client_on_disconnected(struct ADC_client*);
static void ADC_client_on_connected(struct ADC_client*);
static void ADC_client_on_login(struct ADC_client*);
static void ADC_client_connect(struct ADC_client*);
static void ADC_client_disconnect(struct ADC_client*);
static int  ADC_client_create(struct ADC_client* client, int num);
static void ADC_client_destroy(struct ADC_client* client);

static void bot_output(struct ADC_client* client, const char* format, ...)
{
	char logmsg[1024];
	va_list args;
	va_start(args, format);
	vsnprintf(logmsg, 1024, format, args);
	va_end(args);
	if (cfg_debug)
		fprintf(stdout, "* [%4d] %s\n", client->num, logmsg);
}

static void adc_cid_pid(struct ADC_client* client)
{
	char seed[64];
	char pid[64];
	char cid[64];
	uint64_t tiger_res1[3];
	uint64_t tiger_res2[3];

	/* create cid+pid pair */
	memset(seed, 0, 64);
	snprintf(seed, 64, ADCRUSH "%p/%d", client, (int) time(0));
	
	tiger((uint64_t*) seed, strlen(seed), tiger_res1);
	base32_encode((unsigned char*) tiger_res1, TIGERSIZE, pid);
	tiger((uint64_t*) tiger_res1, TIGERSIZE, tiger_res2);
	base32_encode((unsigned char*) tiger_res2, TIGERSIZE, cid);
	
	cid[ADC_CID_SIZE] = 0;
	pid[ADC_CID_SIZE] = 0;
	
	strcat(client->info, " PD");
	strcat(client->info, pid);
	strcat(client->info, " ID");
	strcat(client->info, cid);
}

static size_t get_wait_rand(size_t max)
{
	static size_t next = 0;
	if (next == 0) next = time(0);
	next = next * 1103515245 + 12345;
	return ((size_t )(next / 65536) % max);
}

static void set_state_timeout(struct ADC_client* client, enum protocolState state)
{
	size_t next_timeout = 0;
	
	switch (state)
	{
		case ps_protocol: next_timeout = 30; break;
		case ps_identify: next_timeout = 30; break;
		case ps_verify:   next_timeout = 30; break;
		case ps_normal:   next_timeout = 120; break;
		case ps_none:     next_timeout = 120; break;
	}
	
	if (state == ps_normal || state == ps_none)
	{
		switch (cfg_level)
		{
			case 0: /* polite */
				next_timeout *= 4;
				break;
					
			case 1: /* normal */
				break;
				
			case 2: /* aggressive */
				next_timeout /= 8;
				break;
				
			case 3: /* excessive */
				next_timeout /= 16;
		}
	}
	
	client->state   = state;
	client->timeout = get_wait_rand(next_timeout);
};

static void send_client(struct ADC_client* client, char* msg)
{
	int ret = net_send(client->sd, msg, strlen(msg), UHUB_SEND_SIGNAL);
	
	if (cfg_debug > 1)
	{
		char* dump = strdup(msg);
		dump[strlen(msg) - 1] = 0;
		bot_output(client, "- SEND: '%s'", dump);
		free(dump);
	}
	
 	if (ret != strlen(msg))
	{
		if (ret == -1 && errno != EAGAIN)
		{
			ADC_client_on_disconnected(client);
		}
		else
		{
			printf("ret (%d) != msg->length (%d)\n", ret, (int) strlen(msg));
		}
	}
}


static void ADC_client_on_connected(struct ADC_client* client)
{
	send_client(client, ADC_HANDSHAKE);
	set_state_timeout(client, ps_protocol);
	bot_output(client, "connected.");
}

static void ADC_client_on_disconnected(struct ADC_client* client)
{
	event_del(&client->ev_read);
	event_del(&client->ev_write);
	
	net_close(client->sd);
	client->sd = -1;
	
	bot_output(client, "disconnected.");
	set_state_timeout(client, ps_none);
}

static void ADC_client_on_login(struct ADC_client* client)
{
	bot_output(client, "logged in.");
	set_state_timeout(client, ps_normal);
}


static void send_client_info(struct ADC_client* client)
{
	client->info[0] = 0;
	strcat(client->info, "BINF ");
	strcat(client->info, sid_to_string(client->sid));
	strcat(client->info, " NIadcrush_");
	strcat(client->info, sid_to_string(client->sid));
	strcat(client->info, " VE" ADCRUSH);
	strcat(client->info, " DEcrash\\stest\\sdummy");
	strcat(client->info, " I40.0.0.0");
	strcat(client->info, " EMuhub@extatic.org");
	strcat(client->info, " SL3");
	strcat(client->info, " HN1");
	strcat(client->info, " HR1");
	strcat(client->info, " HO1");
	
	adc_cid_pid(client);
	
	strcat(client->info, "\n");
	
	send_client(client, client->info);
}

static int recv_client(struct ADC_client* client)
{
	ssize_t size = net_recv(client->sd, &client->recvbuf[client->r_offset], ADC_BUFSIZE - client->r_offset, 0);
	if (size == 0 || ((size == -1 && errno != EAGAIN)))
		return -1;
	client->recvbuf[client->r_offset + size] = 0;

	char* start = client->recvbuf;
	char* pos;
	char* lastPos;
	while ((pos = strchr(start, '\n')))
	{
		lastPos = pos;
		pos[0] = 0;
		
		if (cfg_debug > 1)
		{
			bot_output(client, "- RECV: '%s'", start);
		}
		
		fourcc_t cmd = 0;
		if (strlen(start) < 4)
		{
			bot_output(client, "Unexpected response from hub: '%s'", start);
			start = &pos[1];
			continue;
		}
		
		cmd = FOURCC(start[0], start[1], start[2], start[3]);
		
		switch (cmd)
		{
			case ADC_CMD_ISUP:
				break;

			case ADC_CMD_ISID:
				if (client->state == ps_protocol)
				{
					client->sid = string_to_sid(&start[5]);
					client->state = ps_identify;
					send_client_info(client);
					
				}
				break;
				
			case ADC_CMD_IINF:
				break;
				
			case ADC_CMD_BINF:
			{
				if (strlen(start) > 9)
				{
					char t = start[9]; start[9] = 0; sid_t sid = string_to_sid(&start[5]); start[9] = t;
					
					if (sid == client->sid)
					{
						if (client->state == ps_verify || client->state == ps_identify)
						{
							ADC_client_on_login(client);
						}
					}
				}
				break;
			}
			
			case ADC_CMD_ISTA:
				break;
				
			default:
				/* printf("????: '%s'\n", start); */
				break;
		}
		
		start = &pos[1];
	}
	
	client->r_offset = strlen(lastPos);
	memmove(client->recvbuf, lastPos, strlen(lastPos));
	memset(&client->recvbuf[client->r_offset], 0, ADC_BUFSIZE-client->r_offset);
	
	
	return 0;
}

void ADC_client_connect(struct ADC_client* client)
{
	struct timeval timeout = { TIMEOUT_IDLE, 0 };
	net_connect(client->sd, (struct sockaddr*) &saddr, sizeof(struct sockaddr_in));
	event_add(&client->ev_read, &timeout);
	event_add(&client->ev_write, &timeout);
	bot_output(client, "connecting...");
}


void ADC_client_disconnect(struct ADC_client* client)
{
	if (client->sd != -1)
	{
		net_close(client->sd);
		client->sd = -1;
		event_del(&client->ev_read);
		event_del(&client->ev_write);
		bot_output(client, "disconnected.");
	}
}


void event_callback(int fd, short ev, void *arg)
{
	struct ADC_client* client = (struct ADC_client*) arg;

	/* printf("event_callback fd=%d, ev=%d, arg=%p\n", fd, (int) ev, arg); */

	if (ev & EV_WRITE)
	{
		switch (client->state)
		{
			case ps_normal:       break;
			case ps_none:
			{
				int n = 0; 
				char buf[1];
				n = net_recv(fd, buf, 1, MSG_PEEK);
				if (n == 0 || (n == -1 && errno != EAGAIN))
				{
					ADC_client_on_disconnected(client);
				}
				else
				{
					ADC_client_on_connected(client);
				}
				
				break;
			}
			break;
			default:
				break;
		}
	}
	
	if (ev & EV_TIMEOUT)
	{
		if (client->state == ps_none)
		{
			ADC_client_connect(client);
		}
	}
	
	if (ev & EV_READ)
	{
		if (recv_client(client) == -1)
		{
			event_del(&client->ev_read);
		}
	}
}

int ADC_client_create(struct ADC_client* client, int num)
{
	struct timeval timeout = { 0, 0 };
	
	memset(client, 0, sizeof(struct ADC_client));
	client->num = num;
	
	client->sd = net_socket_create(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (client->sd == -1) return -1;
	
	set_state_timeout(client, ps_none);
	
	event_set(&client->ev_write, client->sd, EV_WRITE, event_callback, client);
	event_set(&client->ev_read,  client->sd, EV_READ | EV_PERSIST, event_callback, client);
	
	net_set_nonblocking(client->sd, 1);
	
	timeout.tv_sec = client->timeout;
	evtimer_set(&client->ev_timer, event_callback, client);
	evtimer_add(&client->ev_timer, &timeout);

	return 0;
}

void ADC_client_destroy(struct ADC_client* client)
{
	ADC_client_disconnect(client);
	evtimer_del(&client->ev_timer);
}


void runloop(size_t clients)
{
	size_t n = 0;
	struct ADC_client* client[ADC_MAX_CLIENTS];

	for (n = 0; n < clients; n++)
	{
		struct ADC_client* c = malloc(sizeof(struct ADC_client));
		client[n] = c;

		ADC_client_create(c, n);
		if (n == 0)
		{
			ADC_client_connect(c);
			evtimer_del(&c->ev_timer);
		}
	}
	
	event_dispatch();
	
	for (n = 0; n < clients; n++)
	{
		ADC_client_destroy(client[n]);
		free(client[n]);
	}
}

static void print_version()
{
	printf(ADCRUSH "\n");
	printf("Copyright (c) 2008, Jan Vidar Krey\n");
	printf("\n");
}

static void print_usage(const char* program)
{
	printf("Usage: %s [options] adc://<host>:<port>\n", program);
	
	printf("\n");
	printf("    -a <addr>   Address on the form: adc://<host>:<port>\n");
	printf("    -l <0-3>    Level: 0=polite, 1=normal (default), 2=aggressive, 3=excessive.\n");
	printf("    -b          Bug mode: send buggy commands to the hub.\n");
	printf("    -f          Fuzzer mode: test weird combinations of commands.\n");
	printf("    -s          Security mode: test security policies.\n");
	printf("    -h          Show this help message.\n");
	printf("    -c          Send chat messages.\n");
	printf("    -n <num>    Number of concurrent connections\n");
	printf("    -d          Enable debug (excessive).\n");
	printf("    -q          Quiet mode (no output).\n");
	printf("    -v          Show version number.\n");
	printf("\n");
	
	exit(0);
}

int parse_address(char* arg)
{
	char* split;
	struct hostent* dns;
	struct in_addr* addr;

	if (strlen(arg) < 9)
		return 0;
	
	if (strncmp(arg, "adc://", 6))
		return 0;
	
	split = strrchr(arg+6, ':');
	if (split == 0 || strlen(split) < 2 || strlen(split) > 6)
		return 0;
	
	cfg_port = strtol(split+1, NULL, 10);
	if (cfg_port <= 0 || cfg_port > 65535)
		return 0;
	
	split[0] = 0;

	dns = gethostbyname(arg+6);
	if (dns)
	{
		addr = (struct in_addr*) dns->h_addr_list[0];
		cfg_host = strdup(inet_ntoa(*addr));
	}
	
	if (!cfg_host)
		return 0;
	
	return 1;
}

void parse_command_line(int argc, char** argv)
{
	int opt;
	
	if (argc < 2)
	{
		print_usage(argv[0]);
	}
	
	while ((opt = getopt(argc, argv, "l:n:bfmdqvs")) != -1)
	{
		switch (opt)
		{
			case 'b':
				cfg_bugmode = 1;
				break;

			case 'd':
				cfg_debug++;
				break;

			case 'q':
				cfg_quiet = 1;
				break;

			case 'm':
				cfg_chat = 1;
				break;

			case 'f':
				cfg_fuzz = 1;
				break;

			case 'l':
				if (strlen(optarg) == 1)
				{
					if (optarg[0] == '0')      cfg_level = 0;
					else if (optarg[0] == '1') cfg_level = 1;
					else if (optarg[0] == '2') cfg_level = 2;
					else if (optarg[0] == '3') cfg_level = 3;
					else                       cfg_level = 1;
				}
				else
				{
					cfg_level = 1;
				}
				break;
			
			case 'n':
				cfg_clients = strtol(optarg, NULL, 10);
				if (cfg_clients > ADC_MAX_CLIENTS || cfg_clients == 0) cfg_clients = ADC_CLIENTS_DEFAULT;
				break;		
		
			case 's':
				cfg_security = 1;
				break;

			case 'v':
				print_version();
				break;

			case 'h':
				print_usage(argv[0]);
				break;
	
			default:
				print_usage(argv[0]);
				break;
		}
	}

	if (optind != argc - 1)
	{
		fprintf(stderr, "Address is missing\n");
		print_usage(argv[0]);
	}
	
	if (!parse_address(argv[optind]))
	{
		fprintf(stderr, "Invalid address\n");
		print_usage(argv[0]);
	}
}

int main(int argc, char** argv)
{
	parse_command_line(argc, argv);
	
	net_initialize();
	
	memset(&saddr, 0, sizeof(saddr));
	saddr.sin_family = AF_INET;
	saddr.sin_port   = htons(cfg_port);
	net_string_to_address(AF_INET, cfg_host, &saddr.sin_addr);
	
	runloop(cfg_clients);
	net_shutdown();

	return 0;
}

